package org.dynmap.factions;

import com.massivecraft.factions.entity.Faction;
import com.massivecraft.factions.entity.FactionColl;
import com.massivecraft.factions.entity.MPlayer;
import org.bukkit.configuration.ConfigurationSection;
import org.bukkit.configuration.file.FileConfiguration;
import org.bukkit.plugin.Plugin;
import org.bukkit.plugin.PluginManager;
import org.bukkit.plugin.java.JavaPlugin;
import org.dynmap.DynmapAPI;
import org.dynmap.factions.config.AreaStyle;
import org.dynmap.factions.listener.DynmapFactionsListener;
import org.dynmap.factions.runnable.FactionsUpdate;
import org.dynmap.factions.runnable.PlayerSetUpdate;
import org.dynmap.markers.AreaMarker;
import org.dynmap.markers.Marker;
import org.dynmap.markers.MarkerAPI;
import org.dynmap.markers.MarkerSet;
import org.dynmap.markers.PlayerSet;
import org.jetbrains.annotations.NotNull;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

public class DynmapFactionsPlugin extends JavaPlugin {

    private static Logger log;

    private DynmapAPI api;
    private MarkerAPI markerAPI;
    private boolean playerSets;

    private Map<String, AreaMarker> factionAreaMarkers = new HashMap<>();

    private Map<String, Marker> factionMarkers = new HashMap<>();
    private AreaStyle defaultStyle;
    private Map<String, AreaStyle> customStyle;
    private int blockSize;
    private MarkerSet set;
    private long updatePeriod;

    private Set<String> visible;
    private Set<String> hidden;
    private boolean stop;
    private static DynmapFactionsPlugin instance;
    private boolean reload = false;

    public void setFactionMarkers(Map<String, Marker> factionMarkers) {
        this.factionMarkers = factionMarkers;
    }

    public void setFactionAreaMarkers(Map<String, AreaMarker> factionAreaMarkers) {
        this.factionAreaMarkers = factionAreaMarkers;
    }

    public void setPending(FactionsUpdate pending) {
        this.pending = pending;
    }

    private FactionsUpdate pending = null;

    public MarkerAPI getMarkerAPI() {
        return markerAPI;
    }

    public boolean isPlayerSets() {
        return playerSets;
    }

    public Map<String, AreaMarker> getFactionAreaMarkers() {
        return factionAreaMarkers;
    }

    public Map<String, Marker> getFactionMarkers() {
        return factionMarkers;
    }

    public AreaStyle getDefaultStyle() {
        return defaultStyle;
    }

    public Map<String, AreaStyle> getCustomStyle() {
        return customStyle;
    }

    public int getBlockSize() {
        return blockSize;
    }

    public MarkerSet getSet() {
        return set;
    }

    public long getUpdatePeriod() {
        return updatePeriod;
    }

    public Set<String> getVisible() {
        return visible;
    }

    public Set<String> getHidden() {
        return hidden;
    }

    public boolean isStop() {
        return stop;
    }

    public static DynmapFactionsPlugin getInstance() {
        return instance;
    }

    public FactionsUpdate getPending() {
        return pending;
    }

    /**
     * Logs an info message
     *
     * @param msg <p>The message to log</p>
     */
    public static void info(@NotNull String msg) {
        log.log(Level.INFO, msg);
    }

    public static void severe(String msg) {
        log.log(Level.SEVERE, msg);
    }

    public void requestUpdatePlayerSet(String factionId) {
        if (playerSets) {
            getServer().getScheduler().scheduleSyncDelayedTask(this, new PlayerSetUpdate(this, factionId));
        }
    }

    public void updatePlayerSet(String factionId) {
        /* If Wilderness or other unassociated factions (guid-style ID), skip */
        if (factionId.indexOf('-') >= 0) {
            return;
        }
        Set<String> playerIds = new HashSet<>();
        FactionColl fc = FactionColl.get();

        Faction f = fc.getByName(factionId);    /* Get faction */
        if (f != null) {
            List<MPlayer> ps = f.getMPlayers();
            for (MPlayer fp : ps) {
                playerIds.add(fp.getId());
            }
            factionId = f.getId();
        }
        String setId = "factions." + factionId;
        PlayerSet set = markerAPI.getPlayerSet(setId);  /* See if set exists */
        if ((set == null) && (f != null)) {
            set = markerAPI.createPlayerSet(setId, true, playerIds, false);
            info("Added player visibility set '" + setId + "' for faction " + factionId);
        }

        if (f != null) {
            set.setPlayers(playerIds);
        } else if (set != null) {
            set.deleteSet();
        }
    }

    @Override
    public void onLoad() {
        log = this.getLogger();
    }

    @Override
    public void onEnable() {
        instance = this;
        info("initializing");
        PluginManager pm = getServer().getPluginManager();
        /* Get dynmap */
        Plugin dynmap = pm.getPlugin("dynmap");
        if (dynmap == null) {
            severe("Cannot find dynmap!");
            return;
        }
        api = (DynmapAPI) dynmap; /* Get API */
        /* Get Factions */
        Plugin p = pm.getPlugin("Factions");
        if (p == null) {
            severe("Cannot find Factions!");
            return;
        }

        /* If both enabled, activate */
        if (dynmap.isEnabled() && p.isEnabled()) {
            activate();
        }
    }

    @Override
    public void onDisable() {
        if (set != null) {
            set.deleteMarkerSet();
            set = null;
        }
        factionAreaMarkers.clear();
        stop = true;
    }

    private void updatePlayerSets() {
        if (playerSets) {
            FactionColl fc = FactionColl.get();
            for (Faction f : fc.getAll()) {
                if ((f == fc.getNone()) || (f == fc.getWarzone()) || (f == fc.getSafezone())) {
                    continue;
                }
                updatePlayerSet(f.getId());
            }
        }
    }

    private void activate() {
        markerAPI = api.getMarkerAPI();
        if (markerAPI == null) {
            severe("Error loading dynmap marker API!");
            return;
        }

        blockSize = 16; /* Fixed at 16 */

        /* Load configuration */
        if (reload) {
            this.reloadConfig();
            if (set != null) {
                set.deleteMarkerSet();
                set = null;
            }
        } else {
            reload = true;
        }
        FileConfiguration cfg = getConfig();
        cfg.options().copyDefaults(true);   /* Load defaults, if needed */
        this.saveConfig();  /* Save updates, if needed */

        /* Now, add marker set for mobs (make it transient) */
        set = markerAPI.getMarkerSet("factions.markerSet");
        if (set == null) {
            set = markerAPI.createMarkerSet("factions.markerSet", cfg.getString("layer.name", "Factions"), null, false);
        } else {
            set.setMarkerSetLabel(cfg.getString("layer.name", "Factions"));
        }
        if (set == null) {
            severe("Error creating marker set");
            return;
        }
        /* Make sure these are empty (on reload) */
        factionAreaMarkers.clear();
        factionMarkers.clear();

        int minZoom = cfg.getInt("layer.minZoom", 0);
        if (minZoom > 0) {
            set.setMinZoom(minZoom);
        }
        set.setLayerPriority(cfg.getInt("layer.layerPriority", 10));
        set.setHideByDefault(cfg.getBoolean("layer.hideByDefault", false));


        /* Get style information */
        defaultStyle = new AreaStyle(this, cfg, "regionStyle");
        customStyle = new HashMap<>();
        ConfigurationSection sect = cfg.getConfigurationSection("customStyles");
        if (sect != null) {
            Set<String> ids = sect.getKeys(false);

            for (String id : ids) {
                customStyle.put(id, new AreaStyle(this, cfg, "customStyles." + id, defaultStyle));
            }
        }
        List<String> vis = cfg.getStringList("visibleRegions");
        if (vis != null) {
            visible = new HashSet<>(vis);
        }
        List<String> hid = cfg.getStringList("hiddenRegions");
        if (hid != null) {
            hidden = new HashSet<>(hid);
        }
        /* Check if player sets enabled */
        playerSets = cfg.getBoolean("visibilityByFaction", false);
        if (playerSets) {
            try {
                if (!api.testIfPlayerInfoProtected()) {
                    playerSets = false;
                    info("Dynmap does not have player-info-protected enabled - visibility-by-faction will have no effect");
                }
            } catch (NoSuchMethodError x) {
                playerSets = false;
                info("Dynmap does not support function needed for 'visibility-by-faction' - need to upgrade to 0.60 or later");
            }
        }
        updatePlayerSets();

        /* Set up update job - based on period */
        int per = cfg.getInt("update.period", 300);
        if (per < 15) {
            per = 15;
        }
        updatePeriod = (per * 20L);
        stop = false;

        getServer().getScheduler().scheduleSyncDelayedTask(this, new FactionsUpdate(this, this.getConfig()), 40);   /* First time is 2 seconds */
        getServer().getPluginManager().registerEvents(new DynmapFactionsListener(this), this);

        info("version " + this.getDescription().getVersion() + " is activated");
    }

}
